import pathlib
import tempfile
import unittest
from typing import List
from unittest.mock import patch

from click.testing import CliRunner

from data_pipelines_cli.cli import _cli


class RunTestCommandTestCase(unittest.TestCase):
    commands_to_test = ["run", "test"]
    goldens_dir_path = pathlib.Path(__file__).parent.parent.joinpath("goldens")

    def setUp(self) -> None:
        self.subprocess_run_args = []

    def _mock_run(self, args: List[str]):
        self.subprocess_run_args = args

    @patch(
        "data_pipelines_cli.cli_constants.ENV_CONFIGURATION_PATH",
        goldens_dir_path.joinpath("example_config.yml"),
    )
    def test_no_arg(self):
        for cmd in self.commands_to_test:
            with self.subTest(command=cmd), patch(
                "pathlib.Path.cwd", lambda: self.goldens_dir_path
            ), tempfile.TemporaryDirectory() as tmp_dir, patch(
                "data_pipelines_cli.config_generation.BUILD_DIR", pathlib.Path(tmp_dir)
            ), patch(
                "data_pipelines_cli.cli_commands.compile.BUILD_DIR",
                pathlib.Path(tmp_dir),
            ), patch(
                "data_pipelines_cli.dbt_utils.BUILD_DIR", pathlib.Path(tmp_dir)
            ), patch(
                "data_pipelines_cli.cli_constants.BUILD_DIR", pathlib.Path(tmp_dir)
            ), patch(
                "data_pipelines_cli.dbt_utils.subprocess_run", self._mock_run
            ):
                runner = CliRunner()
                result = runner.invoke(_cli, [cmd])
                self.assertEqual(0, result.exit_code, msg=result.exception)

                self.assertEqual("dbt", self.subprocess_run_args[0])
                self.assertEqual(cmd, self.subprocess_run_args[1])
                args_str = " ".join(self.subprocess_run_args)
                self.assertIn("--profile snowflake", args_str)
                self.assertIn("--target local", args_str)

    @patch(
        "data_pipelines_cli.cli_constants.ENV_CONFIGURATION_PATH",
        goldens_dir_path.joinpath("example_config.yml"),
    )
    def test_dev_arg(self):
        for cmd in self.commands_to_test:
            for env in ["dev", "staging"]:
                with self.subTest(command=cmd, environment=env), patch(
                    "pathlib.Path.cwd", lambda: self.goldens_dir_path
                ), tempfile.TemporaryDirectory() as tmp_dir, patch(
                    "data_pipelines_cli.config_generation.BUILD_DIR",
                    pathlib.Path(tmp_dir),
                ), patch(
                    "data_pipelines_cli.cli_commands.compile.BUILD_DIR",
                    pathlib.Path(tmp_dir),
                ), patch(
                    "data_pipelines_cli.dbt_utils.BUILD_DIR", pathlib.Path(tmp_dir)
                ), patch(
                    "data_pipelines_cli.cli_constants.BUILD_DIR", pathlib.Path(tmp_dir)
                ), patch(
                    "data_pipelines_cli.dbt_utils.subprocess_run", self._mock_run
                ):
                    runner = CliRunner()
                    result = runner.invoke(_cli, [cmd, "--env", env])
                    self.assertEqual(0, result.exit_code, msg=result.exception)

                    self.assertEqual("dbt", self.subprocess_run_args[0])
                    self.assertEqual(cmd, self.subprocess_run_args[1])
                    args_str = " ".join(self.subprocess_run_args)
                    self.assertIn("--profile bigquery", args_str)
                    self.assertIn("--target env_execution", args_str)



from __future__ import annotations

import os
import re
import shutil
import subprocess
import sys
import tempfile
from typing import Optional, Union

import click


# Python's `sed` equivalent, based on the following answer:
# https://stackoverflow.com/a/31499114
def replace(filename: Union[str, os.PathLike[str]], pattern: str, replacement: str) -> None:
    """
    Perform the pure-Python equivalent of in-place `sed` substitution: e.g.,
    ``sed -i -e 's/'${pattern}'/'${replacement}' "${filename}"``.

    Beware however, it uses Python regex dialect instead of `sed`'s one.
    It can introduce regex-related bugs.
    """
    # For efficiency, precompile the passed regular expression.
    pattern_compiled = re.compile(pattern)

    # For portability, NamedTemporaryFile() defaults to mode "w+b" (i.e.,
    # binary writing with updating). This is usually a good thing. In this
    # case, however, binary writing imposes non-trivial encoding constraints
    # trivially resolved by switching to text writing. Let's do that.
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmp_file:
        with open(filename) as src_file:
            for line in src_file:
                tmp_file.write(pattern_compiled.sub(replacement, line))

    # Overwrite the original file with the munged temporary file in a
    # manner preserving file attributes (e.g., permissions).
    shutil.copystat(filename, tmp_file.name)
    shutil.move(tmp_file.name, filename)


def git_revision_hash() -> Optional[str]:
    """
    Get current Git revision hash, if Git is installed and any revision exists.

    :return: Git revision hash, if possible.
    :rtype: Optional[str]
    """
    try:
        rev_process = subprocess.run(["git", "rev-parse", "HEAD"], check=True, capture_output=True)
        return rev_process.stdout.decode("ascii").strip()
    except FileNotFoundError:
        click.echo(
            "The tool has run across an error when trying to get Git "
            "revision hash. Ensure you have `git` installed",
            file=sys.stderr,
        )
        return None
    except subprocess.CalledProcessError as err:
        click.echo(
            "The tool has run across a following error when trying to " "get Git revision hash:",
            file=sys.stderr,
        )
        click.echo(err.stderr, file=sys.stderr)
        return None



import pathlib
import tempfile
import unittest
from unittest.mock import patch

import yaml
from click.testing import CliRunner

from data_pipelines_cli.cli import _cli
from data_pipelines_cli.cli_commands.prepare_env import prepare_env
from data_pipelines_cli.errors import JinjaVarKeyError


class GenHomeProfilesCommandTestCase(unittest.TestCase):
    goldens_dir_path = pathlib.Path(__file__).parent.parent.joinpath("goldens")
    rendered_from_vars_profile = {
        "bigquery": {
            "target": "env_execution",
            "outputs": {
                "env_execution": {
                    "method": "service-account",
                    "project": "example-project",
                    "dataset": "var21-dataset",
                    "keyfile": "/tmp/a/b/c/d.json",
                    "timeout_seconds": 150,
                    "priority": "interactive",
                    "location": "us-west1",
                    "threads": 1337,
                    "retries": 1,
                    "type": "bigquery",
                }
            },
        }
    }

    def setUp(self) -> None:
        self.maxDiff = None

    def test_no_var_profiles_generation(self):
        runner = CliRunner()

        with tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_constants.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch("data_pipelines_cli.config_generation.BUILD_DIR", pathlib.Path(tmp_dir),), patch(
            "data_pipelines_cli.dbt_utils.BUILD_DIR",
            pathlib.Path(tmp_dir),
        ), patch(
            "pathlib.Path.cwd", lambda: self.goldens_dir_path
        ), tempfile.TemporaryDirectory() as tmp_dir2, patch(
            "pathlib.Path.home", lambda: pathlib.Path(tmp_dir2)
        ), patch(
            "data_pipelines_cli.dbt_utils.subprocess_run", lambda _args: None
        ):
            runner.invoke(_cli, ["prepare-env"])
            with open(
                pathlib.Path(tmp_dir2).joinpath(".dbt", "profiles.yml"), "r"
            ) as generated, open(
                self.goldens_dir_path.joinpath("example_profiles", "local_snowflake.yml"),
                "r",
            ) as prepared:
                self.assertDictEqual(yaml.safe_load(prepared), yaml.safe_load(generated))

    def test_vars_profiles_generation(self):
        with tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_constants.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch("data_pipelines_cli.config_generation.BUILD_DIR", pathlib.Path(tmp_dir),), patch(
            "data_pipelines_cli.dbt_utils.BUILD_DIR",
            pathlib.Path(tmp_dir),
        ), patch.dict(
            "os.environ", BIGQUERY_KEYFILE="/tmp/a/b/c/d.json"
        ), patch(
            "pathlib.Path.cwd", lambda: self.goldens_dir_path
        ), tempfile.TemporaryDirectory() as tmp_dir2, patch(
            "pathlib.Path.home", lambda: pathlib.Path(tmp_dir2)
        ), patch(
            "data_pipelines_cli.dbt_utils.subprocess_run", lambda _args: None
        ):
            prepare_env("staging")

            with open(pathlib.Path(tmp_dir2).joinpath(".dbt", "profiles.yml"), "r") as generated:
                self.assertDictEqual(self.rendered_from_vars_profile, yaml.safe_load(generated))

    def test_raise_missing_variable(self):
        with tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_constants.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch("data_pipelines_cli.config_generation.BUILD_DIR", pathlib.Path(tmp_dir),), patch(
            "data_pipelines_cli.cli_commands.prepare_env.read_dbt_vars_from_configs",
            lambda _env: {},
        ), patch.dict(
            "os.environ", BIGQUERY_KEYFILE="/tmp/a/b/c/d.json"
        ), patch(
            "pathlib.Path.cwd", lambda: self.goldens_dir_path
        ), tempfile.TemporaryDirectory() as tmp_dir2, patch(
            "pathlib.Path.home", lambda: pathlib.Path(tmp_dir2)
        ), patch(
            "data_pipelines_cli.dbt_utils.subprocess_run", lambda _args: None
        ):
            with self.assertRaises(JinjaVarKeyError):
                prepare_env("staging")

    def test_raise_missing_environment_variable(self):
        with tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_constants.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch("data_pipelines_cli.config_generation.BUILD_DIR", pathlib.Path(tmp_dir),), patch(
            "data_pipelines_cli.dbt_utils.BUILD_DIR",
            pathlib.Path(tmp_dir),
        ), patch.dict(
            "os.environ", {}
        ), patch(
            "pathlib.Path.cwd", lambda: self.goldens_dir_path
        ), tempfile.TemporaryDirectory() as tmp_dir2, patch(
            "pathlib.Path.home", lambda: pathlib.Path(tmp_dir2)
        ), patch(
            "data_pipelines_cli.dbt_utils.subprocess_run", lambda _args: None
        ):
            with self.assertRaises(JinjaVarKeyError):
                prepare_env("staging")



import pathlib
import unittest
from unittest.mock import patch

from click.testing import CliRunner

from data_pipelines_cli.cli import _cli


class TemplateCommandTestCase(unittest.TestCase):
    example_config_path = pathlib.Path(__file__).parent.parent.joinpath(
        "goldens", "example_config.yml"
    )

    def test_list_templates(self):
        runner = CliRunner()
        with patch(
            "data_pipelines_cli.cli_constants.ENV_CONFIGURATION_PATH",
            self.example_config_path,
        ):
            result = runner.invoke(_cli, ["template-list"])
            self.assertEqual(0, result.exit_code, msg=result.exception)
            self.assertEqual(
                "AVAILABLE TEMPLATES:\n\n"
                "template_name: template1\n"
                "template_path: https://example.com/xyz/abcd.git\n\n"
                "template_name: template2\n"
                "template_path: https://example.com/git/example.git\n\n"
                "template_name: create_test\n"
                "template_path: source_path\n\n",
                result.output,
            )



from typing import Dict, Optional, Sequence

import click
import copier
import questionary

from data_pipelines_cli.cli_utils import echo_warning
from data_pipelines_cli.data_structures import TemplateConfig, read_env_config
from data_pipelines_cli.errors import DataPipelinesError
from data_pipelines_cli.vcs_utils import add_suffix_to_git_template_path


def _choose_template(config_templates: Dict[str, TemplateConfig]) -> TemplateConfig:
    """
    :raises DataPipelinesError: no template found in *config_templates*
    """
    if len(config_templates) == 0:
        raise DataPipelinesError(
            "No template provided. Either run 'dp create <project_path> "
            "<link_to_template>' to use template from the link, or add template "
            "to `~/.dp.yml` file",
        )

    template_name = questionary.select("", choices=list(config_templates.keys())).ask()
    template_config = config_templates[template_name]

    return template_config


def _get_template_path(
    config_templates: Dict[str, TemplateConfig], template_path: Optional[str]
) -> str:
    """:raises DataPipelinesError: no template found in *config_templates*"""
    if template_path:
        if template_path in config_templates.keys():
            to_return = config_templates[template_path]["template_path"]
        else:
            to_return = add_suffix_to_git_template_path(template_path)
    else:
        to_return = _choose_template(config_templates)["template_path"]
    return to_return


def create(project_path: str, template_path: Optional[str]) -> None:
    """
    Create a new project using a template.

    :param project_path: Path to a directory to create
    :type project_path: str
    :param template_path: Path or URI to the repository of the project template
    :type template_path: Optional[str]
    :raises DataPipelinesError: no template found in `.dp.yml` config file
    """
    config = read_env_config()
    config_templates = config["templates"]
    src_template_path = _get_template_path(config_templates, template_path)
    copier.copy(src_path=src_template_path, dst_path=project_path)


@click.command(name="create", help="Create a new project using a template")
@click.argument(
    "project-path",
    type=click.Path(writable=True, path_type=str, dir_okay=True, file_okay=False),
)
@click.argument("template-path", nargs=-1)
def create_command(project_path: str, template_path: Sequence[str]) -> None:
    if template_path and len(template_path) > 1:
        echo_warning("dp create expects at most two arguments -- project-path and template-path")
    create(project_path, template_path[0] if template_path else None)



import pathlib
import sys
from typing import Any, Dict, Tuple

import yaml

from .cli_constants import BUILD_DIR, get_dbt_profiles_env_name
from .cli_utils import echo_subinfo, subprocess_run
from .config_generation import read_dictionary_from_config_directory
from .data_structures import DataPipelinesConfig, read_env_config
from .errors import NoConfigFileError


def read_dbt_vars_from_configs(env: str) -> Dict[str, Any]:
    """Read `vars` field from dp configuration file (``$HOME/.dp.yml``), base
    ``dbt.yml`` config (``config/base/dbt.yml``) and environment-specific config
    (``config/{env}/dbt.yml``) and compile into one dictionary.

    :param env: Name of the environment
    :type env: str
    :return: Dictionary with `vars` and their keys
    :rtype: Dict[str, Any]
    """
    dbt_env_config = read_dictionary_from_config_directory(
        BUILD_DIR.joinpath("dag"), env, "dbt.yml"
    )

    try:
        dp_config = read_env_config()
    except NoConfigFileError:
        dp_config = DataPipelinesConfig(templates={}, vars={})
    dp_vars = dp_config.get("vars", {})
    dbt_vars: Dict[str, str] = dbt_env_config.get("vars", {})

    return dict(dbt_vars, **dp_vars)


def _dump_dbt_vars_from_configs_to_string(env: str) -> str:
    dbt_vars = read_dbt_vars_from_configs(env)
    return yaml.dump(dbt_vars, default_flow_style=True, width=sys.maxsize)


def run_dbt_command(command: Tuple[str, ...], env: str, profiles_path: pathlib.Path) -> None:
    """
    Run dbt subprocess in a context of specified *env*.

    :param command: Tuple representing dbt command and its optional arguments
    :type command: Tuple[str, ...]
    :param env: Name of the environment
    :type env: str
    :param profiles_path: Path to the directory containing `profiles.yml` file
    :type profiles_path: pathlib.Path
    :raises SubprocessNotFound: dbt not installed
    :raises SubprocessNonZeroExitError: dbt exited with error
    """
    command_str = " ".join(list(command))
    echo_subinfo(f"dbt {command_str}")

    dbt_env_config = read_dictionary_from_config_directory(
        BUILD_DIR.joinpath("dag"), env, "dbt.yml"
    )
    dbt_vars = _dump_dbt_vars_from_configs_to_string(env)
    subprocess_run(
        [
            "dbt",
            *command,
            "--profile",
            dbt_env_config["target_type"],
            "--profiles-dir",
            str(profiles_path),
            "--target",
            get_dbt_profiles_env_name(env),
            "--vars",
            dbt_vars,
        ]
    )



import click
import copier
from copier.config.objects import NoSrcPathError

from data_pipelines_cli.cli_utils import echo_warning
from data_pipelines_cli.errors import NotAProjectDirectoryError


def update(project_path: str, vcs_ref: str) -> None:
    """
    Update an existing project from its template
    :param project_path: Path to a directory to create
    :type project_path: str
    :param vcs_ref: Git reference to checkout in projects template
    :type vcs_ref: str
    """
    try:
        copier.copy(dst_path=project_path, vcs_ref=vcs_ref, force=True)
    except NoSrcPathError:
        raise NotAProjectDirectoryError(project_path)


@click.command(name="update", help="Update project from its template")
@click.argument("project-path", nargs=-1)
@click.option("--vcs-ref", default="HEAD", type=str, help="Git reference to checkout")
def update_command(project_path: str, vcs_ref: str) -> None:
    if len(project_path) > 1:
        echo_warning("dp expects at most 1 argument -- project-path")
    update(project_path[0] if project_path else ".", vcs_ref)



class DataPipelinesError(Exception):
    """Base class for all exceptions in data_pipelines_cli module"""

    message: str
    """explanation of the error"""

    def __init__(self, message: str) -> None:
        self.message = message


class DependencyNotInstalledError(DataPipelinesError):
    """Exception raised if certain dependency is not installed"""

    def __init__(self, program_name: str) -> None:
        self.message = (
            f"'{program_name}' not installed. Run 'pip install "
            f"data-pipelines-cli[{program_name}]'"
        )


class NoConfigFileError(DataPipelinesError):
    """Exception raised if `.dp.yml` does not exist"""

    def __init__(self) -> None:
        self.message = "`.dp.yml` config file does not exists. Run 'dp init' to create it."


class NotAProjectDirectoryError(DataPipelinesError):
    """Exception raised if `.copier-answers.yml` file does not exist in given dir"""

    def __init__(self, project_path: str) -> None:
        self.message = (
            f"Given path {project_path} is not a data-pipelines project directory."
            " Run 'dp create' first to create a project"
        )


class SubprocessNonZeroExitError(DataPipelinesError):
    """Exception raised if subprocess exits with non-zero exit code"""

    def __init__(self, subprocess_name: str, exit_code: int) -> None:
        self.message = f"{subprocess_name} has exited with non-zero exit code: {exit_code}"


class SubprocessNotFound(DataPipelinesError):
    """Exception raised if subprocess cannot be found"""

    def __init__(self, subprocess_name: str) -> None:
        self.message = (
            f"{subprocess_name} cannot be found. "
            "Ensure it is installed and listed in your $PATH."
        )


class DockerNotInstalledError(DependencyNotInstalledError):
    """Exception raised if 'docker' is not installed"""

    def __init__(self) -> None:
        super().__init__("docker")


class JinjaVarKeyError(DataPipelinesError):
    def __init__(self, key: str) -> None:
        self.message = (
            f"Variable '{key}' cannot be found neither in 'dbt.yml' and "
            "'$HOME/.dp.yml' vars nor in environment variables, causing Jinja "
            "template rendering to fail."
        )


class AirflowDagsPathKeyError(DataPipelinesError):
    """Exception raised if there is no ``dags_path`` in `airflow.yml` file."""

    def __init__(self) -> None:
        self.message = "Variable 'dags_path' cannot be found in 'airflow.yml' config file."


class DockerErrorResponseError(DataPipelinesError):
    """Exception raised if there is an error response from Docker client."""

    def __init__(self, error_msg: str) -> None:
        self.message = "Error raised when using Docker.\n" + error_msg



import unittest
from unittest.mock import patch

from click.testing import CliRunner

from data_pipelines_cli.cli import _cli
from data_pipelines_cli.cli_commands import update
from data_pipelines_cli.errors import NotAProjectDirectoryError


class UpdateCommandTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.dst_path = None
        self.vcs_ref = ""
        self.copier_force = True

    def _mock_copier(self, dst_path: str, vcs_ref: str, force: bool):
        self.dst_path = dst_path
        self.vcs_ref = vcs_ref
        self.force = force

    def test_update_with_dst_path(self):
        with patch("copier.copy", self._mock_copier):
            runner = CliRunner(mix_stderr=False)
            result = runner.invoke(_cli, ["update", "/some_path/", "/other_path/"])
            self.assertEqual(0, result.exit_code, msg=result.exception)
            self.assertEqual("HEAD", self.vcs_ref)
            self.assertEqual(True, self.force)
            self.assertEqual("/some_path/", self.dst_path)

    def test_update_with_dst_path_and_vcs_ref(self):
        with patch("copier.copy", self._mock_copier):
            runner = CliRunner(mix_stderr=False)
            result = runner.invoke(
                _cli,
                [
                    "update",
                    "/some_path/",
                    "--vcs-ref",
                    "2514ef4ca5929e0f5e7a2b9c702a4cd58a6d2ecf",
                ],
            )
            self.assertEqual(0, result.exit_code, msg=result.exception)
            self.assertEqual("2514ef4ca5929e0f5e7a2b9c702a4cd58a6d2ecf", self.vcs_ref)
            self.assertEqual(True, self.force)
            self.assertEqual("/some_path/", self.dst_path)

    def test_update_no_copier_answers(self):
        with self.assertRaises(NotAProjectDirectoryError):
            update.update("not_existing_path", "HEAD")



import pathlib
from typing import Dict

import click
import yaml

from ..cli_utils import echo_subinfo
from ..config_generation import DbtProfile, generate_profiles_dict
from ..dbt_utils import read_dbt_vars_from_configs, run_dbt_command
from ..jinja import replace_vars_with_values


def prepare_env(env: str) -> None:
    """
    Prepare local environment for use with dbt-related applications.

    Prepare local environment for use with applications expecting a "traditional"
    dbt structure, such as plugins to VS Code. If in doubt, use ``dp run`` and
    ``dp test`` instead.

    :param env: Name of the environment
    :type env: str
    """
    profile: Dict[str, DbtProfile] = replace_vars_with_values(
        generate_profiles_dict(env, True), read_dbt_vars_from_configs(env)
    )

    home_profiles_path = pathlib.Path.home().joinpath(".dbt", "profiles.yml")
    home_profiles_path.parent.mkdir(parents=True, exist_ok=True)
    with open(home_profiles_path, "w") as profiles:
        yaml.dump(profile, profiles, default_flow_style=False)

    echo_subinfo(f"Saved profiles.yml in {home_profiles_path.parent}")
    run_dbt_command(("deps",), env, home_profiles_path.parent)


@click.command(
    name="prepare-env",
    help="Prepare local environment for apps interfacing with dbt",
)
@click.option("--env", default="local", type=str, help="Name of the environment")
def prepare_env_command(env: str) -> None:
    prepare_env(env)



import json
import os
import pathlib
import shutil
import tempfile
import unittest
from typing import List
from unittest.mock import MagicMock, patch

import yaml
from click.testing import CliRunner

from data_pipelines_cli.cli import _cli
from data_pipelines_cli.cli_commands.deploy import DeployCommand
from data_pipelines_cli.errors import (
    AirflowDagsPathKeyError,
    DataPipelinesError,
    DependencyNotInstalledError,
)


def _noop():
    pass


class DeployCommandTestCase(unittest.TestCase):
    provider_args = {
        "schmoken": "/user/schmoken/path",
        "fooser": "user",
        "barsword": "password",
        "answer_to_universe_question": 42,
        "admin_favourite_colour": {
            "r": 255,
            "g": 0,
            "b": 255,
        },
        # `auto_mkdir` is not made up. When passed to LocalFileSystem, it
        # creates directories containing the file if necessary
        "auto_mkdir": True,
    }
    goldens_dir_path = pathlib.Path(__file__).parent.parent.joinpath("goldens")
    dbt_project_config = {"name": "test_project"}

    def setUp(self) -> None:
        blob_json_fd, self.blob_json_filename = tempfile.mkstemp(text=True)
        with open(blob_json_fd, "w") as f:
            yaml.dump(self.provider_args, f)

        self.dbt_project_config_dir = pathlib.Path(tempfile.mkdtemp())
        with open(self.dbt_project_config_dir.joinpath("dbt_project.yml"), "w") as f:
            yaml.dump(self.dbt_project_config, f)

        self.build_temp_dir = pathlib.Path(tempfile.mkdtemp())
        dags_path = pathlib.Path(self.build_temp_dir).joinpath("dag")
        dags_path.mkdir(parents=True)
        shutil.copytree(self.goldens_dir_path.joinpath("config"), dags_path.joinpath("config"))

        self.storage_uri = tempfile.mkdtemp()  # this way fsspec uses LocalFS

        self.subprocess_run_args = []

    def _mock_run(self, args: List[str]):
        self.subprocess_run_args = args

    def tearDown(self) -> None:
        shutil.rmtree(self.storage_uri)
        shutil.rmtree(self.build_temp_dir)
        shutil.rmtree(self.dbt_project_config_dir)
        os.remove(self.blob_json_filename)

    def test_blob_args_types(self):
        for dump, format_name in [(json.dump, "json"), (yaml.dump, "yaml")]:
            with self.subTest(format=format_name):
                runner = CliRunner()

                with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmp_file:
                    dump(self.provider_args, tmp_file)

                result_provider_kwargs = {}

                def mock_init(
                    _self,
                    _env,
                    _docker_push,
                    _blob_address,
                    provider_kwargs_dict,
                    _datahub_ingest,
                ):
                    nonlocal result_provider_kwargs
                    result_provider_kwargs = provider_kwargs_dict

                with patch(
                    "data_pipelines_cli.cli_commands.deploy.DeployCommand.__init__",
                    mock_init,
                ), patch(
                    "data_pipelines_cli.cli_commands.deploy.DeployCommand.deploy",
                    lambda _self: _noop,
                ):
                    result = runner.invoke(_cli, ["deploy", "--blob-args", tmp_file.name])
                self.assertEqual(0, result.exit_code, msg=result.exception)
                self.assertDictEqual(self.provider_args, result_provider_kwargs)

    @patch("data_pipelines_cli.cli_commands.deploy.BUILD_DIR", goldens_dir_path)
    def test_sync_bucket(self):
        # Worth noting: we are not extensively testing
        # 'filesystem_utils.LocalRemoteSync' here. It gets tested in
        # a dedicated 'test_filesystem_utils' file.
        runner = CliRunner()
        with patch("pathlib.Path.cwd", lambda: self.dbt_project_config_dir):
            result = runner.invoke(
                _cli,
                [
                    "deploy",
                    "--dags-path",
                    self.storage_uri,
                    "--blob-args",
                    self.blob_json_filename,
                ],
            )
        self.assertEqual(0, result.exit_code, msg=result.exception)
        self.assertEqual(
            2,
            len(os.listdir(self.storage_uri)),
        )

    def test_no_module_cli(self):
        for module_name, cli_args in [
            ("datahub", ["--datahub-ingest"]),
            ("docker", ["--docker-push"]),
        ]:
            with self.subTest(dep=module_name):
                runner = CliRunner()
                with patch.dict("sys.modules", **{module_name: None}), patch(
                    "pathlib.Path.cwd", lambda: self.dbt_project_config_dir
                ), patch("data_pipelines_cli.cli_constants.BUILD_DIR", self.build_temp_dir):
                    result = runner.invoke(
                        _cli,
                        [
                            "deploy",
                            "--dags-path",
                            self.storage_uri,
                            "--blob-args",
                            self.blob_json_filename,
                            *cli_args,
                        ],
                    )
                self.assertEqual(1, result.exit_code)
                self.assertIsInstance(result.exception, DependencyNotInstalledError)

    def test_no_datahub_method(self):
        with patch.dict("sys.modules", datahub=None), patch(
            "pathlib.Path.cwd", lambda: self.dbt_project_config_dir
        ):
            with self.assertRaises(DependencyNotInstalledError):
                DeployCommand("base", False, self.storage_uri, self.provider_args, True).deploy()

    @patch("data_pipelines_cli.cli_commands.deploy.BUILD_DIR", goldens_dir_path)
    def test_datahub_run(self):
        with patch("pathlib.Path.cwd", lambda: self.dbt_project_config_dir), patch(
            "data_pipelines_cli.cli_commands.deploy.subprocess_run", self._mock_run
        ), patch.dict("sys.modules", datahub=MagicMock()):
            DeployCommand("base", False, self.storage_uri, self.provider_args, True).deploy()
            self.assertListEqual(
                [
                    "datahub",
                    "ingest",
                    "-c",
                    f"{self.goldens_dir_path}/dag/config/base/datahub.yml",
                ],
                self.subprocess_run_args,
            )

    def test_no_docker_method(self):
        with patch.dict("sys.modules", docker=None), patch(
            "pathlib.Path.cwd", lambda: self.dbt_project_config_dir
        ), patch("data_pipelines_cli.cli_constants.BUILD_DIR", self.build_temp_dir):
            with self.assertRaises(DependencyNotInstalledError):
                DeployCommand("base", True, self.storage_uri, self.provider_args, False).deploy()

    @patch(
        "data_pipelines_cli.cli_commands.deploy.BUILD_DIR",
        pathlib.Path("/tmp/some/non/ex/i/sting/path"),
    )
    def test_no_airflow_address(self):
        with self.assertRaises(AirflowDagsPathKeyError):
            DeployCommand("base", False, None, None, False)

    def test_airflow_address(self):
        with tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_commands.deploy.BUILD_DIR", pathlib.Path(tmp_dir)
        ):
            tmp_airflow_path = pathlib.Path(tmp_dir).joinpath(
                "dag", "config", "base", "airflow.yml"
            )
            tmp_airflow_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(
                self.goldens_dir_path.joinpath("config", "base", "airflow.yml"),
                tmp_airflow_path,
            )

            deploy_command = DeployCommand("base", False, None, None, False)
        self.assertEqual(
            "gcs://test-sync-project/sync-dir/dags/my-project-name",
            deploy_command.blob_address_path,
        )

    def test_staging_airflow_address(self):
        with tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_commands.deploy.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.config_generation.BUILD_DIR",
            pathlib.Path(tmp_dir),
        ):
            for env in ["base", "staging"]:
                tmp_config_path = pathlib.Path(tmp_dir).joinpath("dag", "config", env)
                tmp_config_path.mkdir(parents=True, exist_ok=True)
                tmp_file_path = tmp_config_path.joinpath("airflow.yml")
                shutil.copyfile(
                    self.goldens_dir_path.joinpath("config", env, "airflow.yml"),
                    tmp_file_path,
                )

            deploy_command = DeployCommand("staging", False, None, None, False)
        self.assertEqual(
            "gcs://test/jinja/path/com/my/project/name",
            deploy_command.blob_address_path,
        )

    @patch("data_pipelines_cli.cli_commands.deploy.BUILD_DIR", goldens_dir_path)
    def test_docker_run(self):
        docker_kwargs = {}

        def _mock_docker(**kwargs):
            nonlocal docker_kwargs
            docker_kwargs = kwargs
            return []

        docker_images_mock = MagicMock()
        docker_images_mock.configure_mock(**{"push": _mock_docker})
        docker_client_mock = MagicMock()
        docker_client_mock.configure_mock(**{"images": docker_images_mock})
        docker_mock = MagicMock()
        docker_mock.configure_mock(**{"from_env": lambda: docker_client_mock})

        with patch.dict("sys.modules", docker=docker_mock), patch(
            "pathlib.Path.cwd", lambda: self.dbt_project_config_dir
        ), patch("docker.from_env", lambda: docker_client_mock), patch(
            "data_pipelines_cli.data_structures.git_revision_hash", lambda: "sha1234"
        ), patch(
            "data_pipelines_cli.cli_constants.BUILD_DIR", self.build_temp_dir
        ):
            DeployCommand("base", True, self.storage_uri, self.provider_args, False).deploy()

        self.assertEqual("my_docker_repository_uri", docker_kwargs.get("repository"))
        self.assertEqual("sha1234", docker_kwargs.get("tag"))

    @patch("data_pipelines_cli.cli_commands.deploy.BUILD_DIR", goldens_dir_path)
    def test_docker_error(self):
        def _mock_docker(**_kwargs):
            return [
                '{"status":"The push refers to repository [docker.io/library/rep]"}',
                '{"errorDetail":{"message":"An image does not exist locally with '
                'the tag: rep"},"error":"An image does not exist locally with '
                'the tag: rep"}',
            ]

        docker_images_mock = MagicMock()
        docker_images_mock.configure_mock(**{"push": _mock_docker})
        docker_client_mock = MagicMock()
        docker_client_mock.configure_mock(**{"images": docker_images_mock})
        docker_mock = MagicMock()
        docker_mock.configure_mock(**{"from_env": lambda: docker_client_mock})

        with patch("pathlib.Path.cwd", lambda: self.dbt_project_config_dir), patch.dict(
            "sys.modules", docker=docker_mock
        ), patch("data_pipelines_cli.data_structures.git_revision_hash", lambda: "sha1234"), patch(
            "data_pipelines_cli.cli_constants.BUILD_DIR", self.build_temp_dir
        ):
            with self.assertRaises(DataPipelinesError):
                DeployCommand("base", True, self.storage_uri, self.provider_args, False).deploy()



from __future__ import annotations

import os
import pathlib
import shutil
import sys
from typing import Any, Dict, Union

import yaml

from .cli_constants import (
    AVAILABLE_ENVS,
    BUILD_DIR,
    PROFILE_NAME_ENV_EXECUTION,
    get_dbt_profiles_env_name,
)
from .cli_utils import echo_info, echo_subinfo, echo_warning

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict


def _copy_src_dir_to_dst_dir(src_dir: pathlib.Path, dst_dir: pathlib.Path) -> None:
    # It has to be deleted before copying, as `copytree` complains with
    # `FileExistsError`
    if pathlib.Path.exists(dst_dir):
        shutil.rmtree(dst_dir)
    shutil.copytree(src_dir, dst_dir)


def copy_dag_dir_to_build_dir() -> None:
    """Recursively copy `dag` directory to `build/dag` working directory."""
    dag_src_path = pathlib.Path.cwd().joinpath("dag")
    dag_dst_path = BUILD_DIR.joinpath("dag")
    _copy_src_dir_to_dst_dir(dag_src_path, dag_dst_path)


def copy_config_dir_to_build_dir() -> None:
    """Recursively copy `config` directory to `build/dag/config` working directory."""
    config_src_path = pathlib.Path.cwd().joinpath("config")
    dag_dst_path = BUILD_DIR.joinpath("dag", "config")
    echo_info(f"Copying 'config' directory to {dag_dst_path}")
    _copy_src_dir_to_dst_dir(config_src_path, dag_dst_path)


# Heavily based on `config_utils.py` from
# https://github.com/getindata/dbt-airflow-manifest-parser
def read_dictionary_from_config_directory(
    config_path: Union[str, os.PathLike[str]], env: str, file_name: str
) -> Dict[str, Any]:
    """
    Read dictionaries out of *file_name* in both `base` and *env* directories,
    and compile them into one. Values from *env* directory get precedence over
    `base` ones.

    :param config_path: Path to the `config` directory
    :type config_path: Union[str, os.PathLike[str]]
    :param env: Name of the environment
    :type env: str
    :param file_name: Name of the YAML file to parse dictionary from
    :type file_name: str
    :return: Compiled dictionary
    :rtype: Dict[str, Any]
    """
    return dict(
        _read_env_config(config_path, "base", file_name),
        **_read_env_config(config_path, env, file_name),
    )


def _read_env_config(
    config_path: Union[str, os.PathLike[str]], env: str, file_name: str
) -> Dict[str, Any]:
    config_file_path = pathlib.Path(config_path).joinpath("config", env, file_name)
    if config_file_path.exists():
        return _read_yaml_file(config_file_path)
    echo_warning("Missing config file: " + str(config_file_path))
    return {}


def _read_yaml_file(file_path: Union[str, os.PathLike[str]]) -> Dict[str, Any]:
    with open(file_path, "r") as f:
        return yaml.safe_load(f)


class DbtProfile(TypedDict):
    """POD representing dbt's `profiles.yml` file."""

    target: str
    """Name of the `target` for dbt to run"""
    outputs: Dict[str, Dict[str, Any]]
    """Dictionary of a warehouse data and credentials, referenced by `target` name"""


def generate_profiles_dict(env: str, copy_config_dir: bool) -> Dict[str, DbtProfile]:
    """
    Generate and save ``profiles.yml`` file at ``build/profiles/local`` or
    ``build/profiles/env_execution``, depending on `env` argument.

    :param env: Name of the environment
    :type env: str
    :param copy_config_dir: Whether to copy ``config`` directory to ``build`` \
        working directory
    :type copy_config_dir: bool
    :return: Dictionary representing data to be saved in ``profiles.yml``
    :rtype: Dict[str, DbtProfile]
    """
    if copy_config_dir:
        copy_config_dir_to_build_dir()

    dbt_env_config = read_dictionary_from_config_directory(
        BUILD_DIR.joinpath("dag"), env, "dbt.yml"
    )
    dbt_target: str = dbt_env_config["target"]
    dbt_target_type: str = dbt_env_config["target_type"]
    target_type_config = read_dictionary_from_config_directory(
        BUILD_DIR.joinpath("dag"), env, f"{dbt_target_type}.yml"
    )
    target_type_config["type"] = dbt_target_type

    if dbt_target not in AVAILABLE_ENVS:
        echo_warning(
            f"dbt target {dbt_target} is not one of {AVAILABLE_ENVS}. "
            "It can cause errors when running or deploying your project. "
            f"Consider changing target in your 'config/{env}/dbt.yml' to "
            f"{PROFILE_NAME_ENV_EXECUTION}."
        )

    return {
        dbt_target_type: {
            "target": dbt_target,
            "outputs": {dbt_target: target_type_config},
        }
    }


def get_profiles_dir_build_path(env: str) -> pathlib.Path:
    """
    Returns path to ``build/profiles/<profile_name>/``, depending on `env` argument.

    :param env: Name of the environment
    :type env: str
    :return:
    :rtype: pathlib.Path
    """
    profile_name = get_dbt_profiles_env_name(env)
    return BUILD_DIR.joinpath("profiles", profile_name)


def generate_profiles_yml(env: str, copy_config_dir: bool = True) -> pathlib.Path:
    """
    Generate and save ``profiles.yml`` file at ``build/profiles/local`` or
    ``build/profiles/env_execution``, depending on `env` argument.

    :param env: Name of the environment
    :type env: str
    :param copy_config_dir: Whether to copy ``config`` directory to ``build`` \
        working directory
    :type copy_config_dir: bool
    :return: Path to ``build/profiles/{env}``
    :rtype: pathlib.Path
    """
    echo_info("Generating profiles.yml")
    profile = generate_profiles_dict(env, copy_config_dir)

    profiles_path = get_profiles_dir_build_path(env)
    profiles_path.mkdir(parents=True, exist_ok=True)
    with open(profiles_path.joinpath("profiles.yml"), "w") as profiles:
        yaml.dump(profile, profiles, default_flow_style=False)
    echo_subinfo(f"Generated profiles.yml in {profiles_path}")

    return profiles_path



import pathlib
import tempfile
import unittest
from unittest.mock import MagicMock, patch

import yaml
from click.testing import CliRunner

from data_pipelines_cli.cli import _cli
from data_pipelines_cli.errors import DataPipelinesError, NoConfigFileError


class CreateCommandTestCase(unittest.TestCase):
    copier_src_path = "source_path"
    copier_dst_path = "destination_path"
    goldens_dir_path = pathlib.Path(__file__).parent.parent.joinpath("goldens")

    def _mock_copier(self, src_path: str, dst_path: str):
        self.assertEqual(self.copier_src_path, src_path)
        self.assertEqual(self.copier_dst_path, dst_path)

    def test_create_no_config(self):
        with tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_constants.ENV_CONFIGURATION_PATH",
            pathlib.Path(tmp_dir).joinpath("non_existing_file"),
        ):
            runner = CliRunner(mix_stderr=False)
            result = runner.invoke(_cli, ["create", "some_path"])
            self.assertEqual(1, result.exit_code)
            self.assertIsInstance(result.exception, NoConfigFileError)

    @patch(
        "data_pipelines_cli.cli_constants.ENV_CONFIGURATION_PATH",
        goldens_dir_path.joinpath("example_config.yml"),
    )
    def test_create_with_template_path(self):
        with patch("copier.copy", self._mock_copier):
            runner = CliRunner(mix_stderr=False)
            result = runner.invoke(_cli, ["create", self.copier_dst_path, self.copier_src_path])
            self.assertEqual(0, result.exit_code, msg=result.exception)

    @patch(
        "data_pipelines_cli.cli_constants.ENV_CONFIGURATION_PATH",
        goldens_dir_path.joinpath("example_config.yml"),
    )
    def test_create_with_template_name(self):
        with patch("copier.copy", self._mock_copier):
            runner = CliRunner(mix_stderr=False)
            result = runner.invoke(_cli, ["create", self.copier_dst_path, "create_test"])
            self.assertEqual(0, result.exit_code, msg=result.exception)

    @patch(
        "data_pipelines_cli.cli_constants.ENV_CONFIGURATION_PATH",
        goldens_dir_path.joinpath("example_config.yml"),
    )
    @patch("questionary.select")
    def test_create_without_template_path(self, mock_select):
        magic_mock = MagicMock()
        magic_mock.configure_mock(**{"ask": lambda: "create_test"})
        mock_select.return_value = magic_mock

        with patch("copier.copy", self._mock_copier):
            runner = CliRunner(mix_stderr=False)
            result = runner.invoke(_cli, ["create", self.copier_dst_path])
            self.assertEqual(0, result.exit_code, msg=result.exception)

    def test_no_templates(self):
        with tempfile.NamedTemporaryFile() as tmp_file, patch(
            "data_pipelines_cli.cli_constants.ENV_CONFIGURATION_PATH",
            pathlib.Path(tmp_file.name),
        ):
            with open(tmp_file.name, "w") as f:
                yaml.dump({"vars": {}, "templates": {}}, f)
            runner = CliRunner(mix_stderr=False)
            result = runner.invoke(_cli, ["create", "some_path"])
            self.assertEqual(1, result.exit_code)
            self.assertIsInstance(result.exception, DataPipelinesError)
            self.assertRegex(result.exception.message, r"^No template provided\..*")



import json
import pathlib
import shutil
import tempfile
import unittest
from os import PathLike
from typing import Any
from unittest.mock import MagicMock, patch

import yaml
from click.testing import CliRunner

from data_pipelines_cli.cli import _cli
from data_pipelines_cli.cli_commands.publish import create_package
from data_pipelines_cli.errors import DataPipelinesError

goldens_dir_path = pathlib.Path(__file__).parent.parent.joinpath("goldens")


class PublishCommandTestCase(unittest.TestCase):
    expected_sources = {
        "version": 2,
        "sources": [
            {
                "name": "my_test_project_1337",
                "database": "example-project",
                "schema": "username_private_working_dataset",
                "tags": ["project:my_test_project_1337"],
                "meta": {"dag": "experimental-dag"},
                "tables": [
                    {
                        "name": "my_first_dbt_model",
                        "description": "A starter dbt model",
                        "columns": [
                            {
                                "name": "id",
                                "description": "The primary key for this table",
                                "meta": {},
                                "quote": None,
                                "tags": [],
                            }
                        ],
                        "meta": {},
                        "tags": [],
                    },
                    {
                        "name": "my_second_dbt_model",
                        "description": "A starter dbt model",
                        "columns": [
                            {
                                "name": "id",
                                "description": "The primary key for this table",
                                "meta": {},
                                "quote": None,
                                "tags": [],
                            }
                        ],
                        "meta": {},
                        "tags": [],
                    },
                ],
            }
        ],
    }
    dbt_project = {
        "config-version": 2,
        "name": "my_test_project_1337_sources",
        "version": "1.2.3",
        "source-paths": ["models"],
    }

    def setUp(self) -> None:
        self.maxDiff = None

        self.build_temp_dir = pathlib.Path(tempfile.mkdtemp())
        dags_path = pathlib.Path(self.build_temp_dir).joinpath("dag")
        dags_path.mkdir(parents=True)
        shutil.copytree(goldens_dir_path.joinpath("config"), dags_path.joinpath("config"))

        profiles_yml_path = pathlib.Path(self.build_temp_dir).joinpath(
            "profiles", "env_execution", "profiles.yml"
        )
        profiles_yml_path.parent.mkdir(parents=True)
        shutil.copyfile(
            goldens_dir_path.joinpath("example_profiles", "dev_bigquery.yml"),
            profiles_yml_path,
        )

    def tearDown(self) -> None:
        shutil.rmtree(self.build_temp_dir)

    def mock_origin(self, name: str):
        self.origin = MagicMock()
        self.origin.push = MagicMock()
        return self.origin

    def mock_clone_from(self, url: PathLike, to_path: PathLike, **kwargs: Any):
        self.assertEqual("https://gitlab.com/getindata/dataops/some_repo.git", url)
        self.assertEqual("main", kwargs["branch"])

        def noop():
            pass

        repo_mock = MagicMock()
        self.git = MagicMock()
        self.index = MagicMock()
        self.index.commit = MagicMock()
        self.git.add = MagicMock()
        config_writer_mock = MagicMock()
        set_value_mock = MagicMock()
        set_value_mock.configure_mock(**{"release": noop})
        config_writer_mock.configure_mock(**{"set_value": lambda x, y, z: set_value_mock})
        repo_mock.configure_mock(
            **{
                "config_writer": config_writer_mock,
                "git": self.git,
                "index": self.index,
                "remote": self.mock_origin,
            }
        )
        return repo_mock

    def repo_class_mock(self):
        self.repo_mock_class = MagicMock()
        self.repo_mock_class.configure_mock(**{"clone_from": self.mock_clone_from})
        return self.repo_mock_class

    @patch("pathlib.Path.cwd", lambda: goldens_dir_path)
    def test_generate_correct_project(self):
        runner = CliRunner()
        with patch("data_pipelines_cli.cli_commands.publish.BUILD_DIR", self.build_temp_dir), patch(
            "data_pipelines_cli.config_generation.BUILD_DIR", self.build_temp_dir
        ), patch("data_pipelines_cli.cli_commands.publish.Repo", self.repo_class_mock()):
            runner.invoke(_cli, ["publish", "--key-path", "SOME_KEY.txt"])
            result = runner.invoke(_cli, ["publish", "--key-path", "SOME_KEY.txt"])

            self.verify_status_code(result)
            self.verify_publications()
            self.verify_generated_files()

    def verify_status_code(self, result):
        self.assertEqual(0, result.exit_code, msg=result.output)

    def verify_publications(self):
        self.git.add.assert_called_with(all=True)
        self.index.commit.assert_called_with(
            "Publication from project my_test_project_1337, version: 1.2.3"
        )
        self.origin.push.assert_called_with()

    def verify_generated_files(self):
        with open(
            pathlib.Path(self.build_temp_dir).joinpath("package", "models", "sources.yml"),
            "r",
        ) as sources_yml:
            self.assertDictEqual(self.expected_sources, yaml.safe_load(sources_yml))
        with open(
            pathlib.Path(self.build_temp_dir).joinpath("package", "dbt_project.yml"),
            "r",
        ) as dbt_project_yml:
            self.assertDictEqual(self.dbt_project, yaml.safe_load(dbt_project_yml))

    def test_no_models(self):
        with tempfile.TemporaryDirectory() as tmp_dir, patch(
            "pathlib.Path.cwd", lambda: pathlib.Path(tmp_dir)
        ):
            shutil.copyfile(
                goldens_dir_path.joinpath("dbt_project.yml"),
                pathlib.Path(tmp_dir).joinpath("dbt_project.yml"),
            )

            target_path = pathlib.Path(tmp_dir).joinpath("target")
            target_path.mkdir(parents=True)
            with open(
                goldens_dir_path.joinpath("target", "manifest.json"), "r"
            ) as manifest_json, open(target_path.joinpath("manifest.json"), "w") as tmp_manifest:
                manifest = json.load(manifest_json)
                for k in list(manifest["nodes"].keys()):
                    if k.startswith("model"):
                        del manifest["nodes"][k]
                json.dump(manifest, tmp_manifest)
            with self.assertRaises(DataPipelinesError):
                create_package()



import json
from typing import Dict, Iterable, List, Optional, Union, cast

import click

from data_pipelines_cli.errors import DockerErrorResponseError


class DockerReadResponse:
    """POD representing Docker response processed by :class:`DockerResponseReader`."""

    msg: str
    """Read and processed message"""
    is_error: bool
    """Whether response is error or not"""

    def __init__(self, msg: str, is_error: bool) -> None:
        self.msg = msg
        self.is_error = is_error

    def __str__(self) -> str:
        return self.msg


class DockerResponseReader:
    """
    Read and process Docker response.

    Docker response turns into processed strings instead of plain dictionaries.
    """

    logs_generator: Iterable[Union[str, Dict[str, Union[str, Dict[str, str]]]]]
    """Iterable representing Docker response"""
    cached_read_response: Optional[List[DockerReadResponse]]
    """Internal cache of already processed response"""

    def __init__(
        self,
        logs_generator: Iterable[Union[str, Dict[str, Union[str, Dict[str, str]]]]],
    ):
        self.logs_generator = logs_generator
        self.cached_read_response = None

    def read_response(self) -> List[DockerReadResponse]:
        """
        Read and process Docker response.

        :return: List of processed lines of response
        :rtype: List[DockerReadResponse]
        """
        to_return = []

        for log in self.logs_generator:
            if isinstance(log, str):
                log = json.loads(log)
            log = cast(Dict[str, Union[str, Dict[str, str]]], log)

            if "status" in log:
                to_return.append(self._prepare_status(log))
            if "stream" in log:
                to_return += self._prepare_stream(log)
            if "aux" in log:
                to_return += self._prepare_aux(log)

            if "errorDetail" in log:
                to_return.append(self._prepare_error_detail(log))
            elif "error" in log:
                to_return.append(self._prepare_error(log))

        self.cached_read_response = to_return
        return to_return

    def click_echo_ok_responses(self) -> None:
        """Read, process and print positive Docker updates.

        :raises DockerErrorResponseError: Came across error update in Docker response.
        """
        read_response = self.cached_read_response or self.read_response()

        for response in read_response:
            if response.is_error:
                raise DockerErrorResponseError(response.msg)
            click.echo(response.msg)

    @staticmethod
    def _prepare_status(log: Dict[str, Union[str, Dict[str, str]]]) -> DockerReadResponse:
        status_message = cast(str, log["status"])
        progress_detail = cast(str, log.get("progressDetail", ""))
        status_id = cast(str, log.get("id", ""))
        message = (
            status_message
            + (f" ({status_id})" if status_id else "")
            + (f": {progress_detail}" if progress_detail else "")
        )

        return DockerReadResponse(message, False)

    @staticmethod
    def _prepare_stream(log: Dict[str, Union[str, Dict[str, str]]]) -> List[DockerReadResponse]:
        stream = cast(str, log["stream"])
        return list(
            map(
                lambda line: DockerReadResponse(line, False),
                filter(lambda x: x, stream.splitlines()),
            )
        )

    @staticmethod
    def _prepare_aux(log: Dict[str, Union[str, Dict[str, str]]]) -> List[DockerReadResponse]:
        aux = cast(Dict[str, str], log["aux"])
        to_return = []
        if "Digest" in aux:
            to_return.append(DockerReadResponse(f"Digest: {aux['Digest']}", False))
        if "ID" in aux:
            to_return.append(DockerReadResponse(f"ID: {aux['ID']}", False))
        return to_return

    @staticmethod
    def _prepare_error_detail(log: Dict[str, Union[str, Dict[str, str]]]) -> DockerReadResponse:
        error_detail = cast(Dict[str, str], log["errorDetail"])
        error_message = error_detail.get("message", "")
        error_code = error_detail.get("code", None)
        return DockerReadResponse(
            "ERROR: " + error_message + (f"\nError code: {error_code}" if error_code else ""),
            True,
        )

    @staticmethod
    def _prepare_error(log: Dict[str, Union[str, Dict[str, str]]]) -> DockerReadResponse:
        return DockerReadResponse("ERROR: " + cast(str, log["error"]), True)



import pathlib
import tempfile
from typing import Optional, Sequence

import click
import copier
import questionary
import yaml

from ..cli_constants import DEFAULT_GLOBAL_CONFIG, ENV_CONFIGURATION_PATH
from ..data_structures import DataPipelinesConfig
from ..errors import DataPipelinesError
from ..vcs_utils import add_suffix_to_git_template_path


def _download_global_config(config_path: str) -> DataPipelinesConfig:
    config_path = add_suffix_to_git_template_path(config_path)

    with tempfile.TemporaryDirectory() as tmp:
        copier.copy(config_path, tmp, quiet=True)
        with open(pathlib.Path(tmp).joinpath("dp.yml")) as config_file:
            config = yaml.safe_load(config_file)
    return config


def init(config_path: Optional[str]) -> None:
    """
    Configure the tool for the first time.

    :param config_path: URI of the repository with a template of the config file
    :type config_path: Optional[str]
    :raises DataPipelinesError: user do not want to overwrite existing config file
    """
    if ENV_CONFIGURATION_PATH.is_file():
        overwrite_confirm = questionary.confirm(
            "dp config already exists. Do you want to overwrite it?",
            default=False,
        ).ask()
        if not overwrite_confirm:
            raise DataPipelinesError("Could not overwrite existing config")

    if config_path:
        config = _download_global_config(config_path)
    else:
        config = DEFAULT_GLOBAL_CONFIG

    with open(ENV_CONFIGURATION_PATH, "w") as f:
        yaml.dump(config, f, default_flow_style=False)


@click.command(name="init", help="Configure the tool for the first time")
@click.argument("config_path", nargs=-1)
def init_command(config_path: Sequence[str]) -> None:
    init(config_path[0] if config_path else None)



import sys

import click

from .cli_commands.clean import clean_command
from .cli_commands.compile import compile_project_command
from .cli_commands.create import create_command
from .cli_commands.deploy import deploy_command
from .cli_commands.init import init_command
from .cli_commands.prepare_env import prepare_env_command
from .cli_commands.publish import publish_command
from .cli_commands.run import run_command
from .cli_commands.template import list_templates_command
from .cli_commands.test import test_command
from .cli_commands.update import update_command
from .cli_utils import echo_error
from .errors import DataPipelinesError


@click.group()
@click.version_option(prog_name="dp")
def _cli() -> None:
    pass


def cli() -> None:
    try:
        _cli()
    except DataPipelinesError as err:
        echo_error(f"CLI Error: {err.message}")
        sys.exit(1)


_cli.add_command(clean_command)
_cli.add_command(compile_project_command)
_cli.add_command(create_command)
_cli.add_command(deploy_command)
_cli.add_command(init_command)
_cli.add_command(prepare_env_command)
_cli.add_command(publish_command)
_cli.add_command(run_command)
_cli.add_command(list_templates_command)
_cli.add_command(test_command)
_cli.add_command(update_command)



import json
import pathlib
import shutil
from typing import Any, Dict, List, Tuple, cast

import click
import yaml
from dbt.contracts.graph.manifest import Manifest, ManifestNode
from dbt.contracts.graph.parsed import ColumnInfo
from git import Repo

from ..cli_constants import BUILD_DIR
from ..cli_utils import echo_info
from ..config_generation import read_dictionary_from_config_directory
from ..data_structures import DbtModel, DbtSource, DbtTableColumn
from ..errors import DataPipelinesError


def _get_project_name_and_version() -> Tuple[str, str]:
    with open(pathlib.Path.cwd().joinpath("dbt_project.yml"), "r") as f:
        dbt_project_config = yaml.safe_load(f)
        return dbt_project_config["name"], dbt_project_config["version"]


def _get_database_and_schema_name(manifest: Manifest) -> Tuple[str, str]:
    try:
        model = next(
            node
            for node in map(lambda n: cast(ManifestNode, n), manifest.nodes.values())
            if node.resource_type == "model"
        )
        return model.database, model.schema
    except StopIteration:
        raise DataPipelinesError("There is no model in 'manifest.json' file.")


def _parse_columns_dict_into_table_list(columns: Dict[str, ColumnInfo]) -> List[DbtTableColumn]:
    return [
        DbtTableColumn(
            name=column.name,
            description=column.description,
            meta=column.meta,
            quote=column.quote,
            tags=column.tags,
        )
        for column in columns.values()
    ]


def _parse_models_schema(manifest: Manifest) -> List[DbtModel]:
    return [
        DbtModel(
            name=node.name,
            description=node.description,
            tags=node.tags,
            meta=node.meta,
            columns=_parse_columns_dict_into_table_list(node.columns),
        )
        for node in map(lambda n: cast(ManifestNode, n), manifest.nodes.values())
        if node.resource_type == "model"
    ]


def _get_dag_id() -> str:
    with open(BUILD_DIR.joinpath("dag", "config", "base", "airflow.yml"), "r") as airflow_yml:
        return yaml.safe_load(airflow_yml)["dag"]["dag_id"]


def _create_source(project_name: str) -> DbtSource:
    with open(pathlib.Path.cwd().joinpath("target", "manifest.json"), "r") as manifest_json:
        manifest_dict = json.load(manifest_json)
        manifest = Manifest.from_dict(manifest_dict)

    database_name, schema_name = _get_database_and_schema_name(manifest)

    return DbtSource(
        name=project_name,
        database=database_name,
        schema=schema_name,
        tables=_parse_models_schema(manifest),
        meta={"dag": _get_dag_id()},
        tags=[f"project:{project_name}"],
    )


def _create_dbt_project(project_name: str, project_version: str) -> Dict[str, Any]:
    return {
        "name": f"{project_name}_sources",
        "source-paths": ["models"],
        "version": project_version,
        "config-version": 2,
    }


def create_package() -> pathlib.Path:
    """Create a dbt package out of the built project.

    :raises DataPipelinesError: There is no model in 'manifest.json' file.
    """
    project_name, project_version = _get_project_name_and_version()
    package_path = BUILD_DIR.joinpath("package")

    sources_path = package_path.joinpath("models", "sources.yml")
    sources_path.parent.mkdir(parents=True, exist_ok=True)
    with open(sources_path, "w") as sources_yml:
        yaml.dump(
            {"version": 2, "sources": [_create_source(project_name)]},
            sources_yml,
            default_flow_style=False,
        )

    with open(package_path.joinpath("dbt_project.yml"), "w") as dbt_project_yml:
        yaml.dump(
            _create_dbt_project(project_name, project_version),
            dbt_project_yml,
            default_flow_style=False,
        )
    return package_path


def _clean_repo(packages_repo: pathlib.Path) -> None:
    if packages_repo.exists():
        echo_info(f"Removing {packages_repo}")
        shutil.rmtree(packages_repo)


def _copy_publication_to_repo(package_dest: pathlib.Path, package_path: pathlib.Path) -> None:
    if package_dest.exists():
        echo_info(f"Removing {package_dest}")
        shutil.rmtree(package_dest)
    echo_info("Copying new version")
    shutil.copytree(package_path, package_dest)


def _configure_git_env(repo: Repo, config: Dict[str, Any]) -> None:
    repo.config_writer().set_value("user", "name", config["username"]).release()
    repo.config_writer().set_value("user", "email", config["email"]).release()


def _commit_and_push_changes(repo: Repo, project_name: str, project_version: str) -> None:
    echo_info("Publishing")
    repo.git.add(all=True)
    repo.index.commit(f"Publication from project {project_name}, version: {project_version}")
    origin = repo.remote(name="origin")
    origin.push()


def publish_package(package_path: pathlib.Path, key_path: str, env: str) -> None:
    packages_repo = BUILD_DIR.joinpath("packages_repo")
    publish_config = read_dictionary_from_config_directory(
        BUILD_DIR.joinpath("dag"), env, "publish.yml"
    )
    _clean_repo(packages_repo)
    ssh_command_with_key = f"ssh -i {key_path}"
    repo = Repo.clone_from(
        publish_config["repository"],
        packages_repo,
        branch=publish_config["branch"],
        env={"GIT_SSH_COMMAND": ssh_command_with_key},
    )
    with repo.git.custom_environment(GIT_SSH_COMMAND=ssh_command_with_key):
        project_name, project_version = _get_project_name_and_version()
        _copy_publication_to_repo(packages_repo.joinpath(project_name), package_path)
        _configure_git_env(repo, publish_config)
        _commit_and_push_changes(repo, project_name, project_version)


@click.command(name="publish", help="Create a dbt package out of the project")
@click.option(
    "--key-path",
    type=str,
    required=True,
    help="Path to the key with write access to repo with published packages",
)
@click.option(
    "--env",
    default="base",
    type=str,
    show_default=True,
    required=True,
    help="Name of the environment",
)
def publish_command(key_path: str, env: str) -> None:
    package_path = create_package()
    publish_package(package_path, key_path, env)



import pathlib
import tempfile
import unittest
from unittest.mock import MagicMock, patch

import yaml
from click.testing import CliRunner

from data_pipelines_cli.cli import _cli
from data_pipelines_cli.cli_commands.init import init
from data_pipelines_cli.cli_constants import DEFAULT_GLOBAL_CONFIG
from data_pipelines_cli.data_structures import (
    DataPipelinesConfig,
    TemplateConfig,
    read_env_config,
)
from data_pipelines_cli.errors import DataPipelinesError


class InitCommandTestCase(unittest.TestCase):
    test_config_template_path = pathlib.Path(__file__).parent.parent.joinpath(
        "goldens", "config_template"
    )
    example_config_dict = DataPipelinesConfig(
        templates={
            "my-template": TemplateConfig(
                template_name="my-template",
                template_path="https://example.com/git/example.git",
            ),
            "local-template": TemplateConfig(
                template_name="local-template",
                template_path="/var/tmp/Documents/project-template",
            ),
        },
        vars={
            "username": "test_user",
        },
    )

    def test_init(self):
        runner = CliRunner()
        with tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_commands.init.ENV_CONFIGURATION_PATH",
            pathlib.Path(tmp_dir).joinpath(".dp.yml"),
        ), patch(
            "data_pipelines_cli.cli_constants.ENV_CONFIGURATION_PATH",
            pathlib.Path(tmp_dir).joinpath(".dp.yml"),
        ):
            result = runner.invoke(
                _cli,
                ["init", str(self.test_config_template_path)],
                input="test_user\n/var/tmp",
            )
            self.assertEqual(0, result.exit_code, msg=result.exception)
            self.assertEqual(self.example_config_dict, read_env_config())

    def test_global_config(self):
        runner = CliRunner()
        with tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_commands.init.ENV_CONFIGURATION_PATH",
            pathlib.Path(tmp_dir).joinpath(".dp.yml"),
        ), patch(
            "data_pipelines_cli.cli_constants.ENV_CONFIGURATION_PATH",
            pathlib.Path(tmp_dir).joinpath(".dp.yml"),
        ):
            result = runner.invoke(_cli, ["init"])
            self.assertEqual(0, result.exit_code, msg=result.exception)
            self.assertEqual(DEFAULT_GLOBAL_CONFIG, read_env_config())

    @patch("questionary.confirm")
    def test_overwrite_yes(self, mock_questionary):
        magic_mock = MagicMock()
        magic_mock.configure_mock(**{"ask": lambda: True})
        mock_questionary.return_value = magic_mock

        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_dp_path = pathlib.Path(tmp_dir).joinpath(".dp.yml")
            with patch(
                "data_pipelines_cli.cli_commands.init.ENV_CONFIGURATION_PATH",
                tmp_dp_path,
            ), patch(
                "data_pipelines_cli.cli_constants.ENV_CONFIGURATION_PATH",
                tmp_dp_path,
            ):
                with open(tmp_dp_path, "w") as tmp_file:
                    yaml.dump(self.example_config_dict, tmp_file)
                    self.assertEqual(self.example_config_dict, read_env_config())
                init(None)
                self.assertEqual(DEFAULT_GLOBAL_CONFIG, read_env_config())

    @patch("questionary.confirm")
    def test_overwrite_no(self, mock_questionary):
        magic_mock = MagicMock()
        magic_mock.configure_mock(**{"ask": lambda: False})
        mock_questionary.return_value = magic_mock

        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_dp_path = pathlib.Path(tmp_dir).joinpath(".dp.yml")
            with patch(
                "data_pipelines_cli.cli_commands.init.ENV_CONFIGURATION_PATH",
                tmp_dp_path,
            ), patch(
                "data_pipelines_cli.cli_constants.ENV_CONFIGURATION_PATH",
                tmp_dp_path,
            ):
                with open(tmp_dp_path, "w") as tmp_file:
                    yaml.dump(self.example_config_dict, tmp_file)
                    self.assertEqual(self.example_config_dict, read_env_config())

                with self.assertRaises(DataPipelinesError):
                    init(None)
                self.assertEqual(self.example_config_dict, read_env_config())



from __future__ import annotations

import os
import subprocess
import sys
from typing import Any, List, Optional

import click

from data_pipelines_cli.errors import (
    DataPipelinesError,
    SubprocessNonZeroExitError,
    SubprocessNotFound,
)


def echo_error(text: str, **kwargs: Any) -> None:
    """
    Print an error message to stderr using click-specific print function.

    :param text: Message to print
    :type text: str
    :param kwargs:
    """
    click.secho(text, file=sys.stderr, fg="red", **kwargs)


def echo_suberror(text: str, **kwargs: Any) -> None:
    """
    Print a suberror message to stderr using click-specific print function.

    :param text: Message to print
    :type text: str
    :param kwargs:
    """
    click.secho(text, file=sys.stderr, fg="bright_red", **kwargs)


def echo_warning(text: str, **kwargs: Any) -> None:
    """
    Print a warning message to stderr using click-specific print function.

    :param text: Message to print
    :type text: str
    :param kwargs:
    """
    click.secho(text, file=sys.stderr, fg="yellow", **kwargs)


def echo_info(text: str, **kwargs: Any) -> None:
    """
    Print a message to stdout using click-specific print function.

    :param text: Message to print
    :type text: str
    :param kwargs:
    """
    click.secho(text, fg="blue", bold=True, **kwargs)


def echo_subinfo(text: str, **kwargs: Any) -> None:
    """
    Print a subinfo message to stdout using click-specific print function.

    :param text: Message to print
    :type text: str
    :param kwargs:
    """
    click.secho(text, fg="bright_blue", **kwargs)


def get_argument_or_environment_variable(
    argument: Optional[str], argument_name: str, environment_variable_name: str
) -> str:
    """
    Given *argument* is not ``None``, return its value. Otherwise, search
    for *environment_variable_name* amongst environment variables and return
    it. If such a variable is not set, raise :exc:`.DataPipelinesError`.

    :param argument: Optional value passed to the CLI as the *argument_name*
    :type argument: Optional[str]
    :param argument_name: Name of the CLI's argument
    :type argument_name: str
    :param environment_variable_name: Name of the environment variable to search for
    :type environment_variable_name: str
    :return: Value of the *argument* or specified environment variable
    :raises DataPipelinesError: *argument* is ``None`` and \
        *environment_variable_name* is not set
    """
    result = argument or os.environ.get(environment_variable_name)
    if not result:
        raise DataPipelinesError(
            f"Could not get {environment_variable_name}. Either set it as an "
            f"environment variable {environment_variable_name} or pass as a "
            f"`--{argument_name}` CLI argument."
        )
    return result


def subprocess_run(args: List[str]) -> subprocess.CompletedProcess[bytes]:
    """
    Run subprocess and return its state if completed with a success. If not,
    raise :exc:`.SubprocessNonZeroExitError`.

    :param args: List of strings representing subprocess and its arguments
    :type args: List[str]
    :return: State of the completed process
    :rtype: subprocess.CompletedProcess[bytes]
    :raises SubprocessNonZeroExitError: subprocess exited with non-zero exit code
    """
    try:
        return subprocess.run(args, check=True)
    except FileNotFoundError:
        raise SubprocessNotFound(args[0])
    except subprocess.CalledProcessError as err:
        raise SubprocessNonZeroExitError(args[0], err.returncode)



import click
import yaml

from data_pipelines_cli.data_structures import read_env_config


def list_templates() -> None:
    """Print a list of all templates saved in the config file."""
    config = read_env_config()

    click.echo("AVAILABLE TEMPLATES:\n")
    for tc in config["templates"].values():
        click.echo(yaml.dump(tc))


@click.command(name="template-list", help="Print a list of all templates saved in the config file")
def list_templates_command() -> None:
    list_templates()



import json
import os
import pathlib
import shutil
import tempfile
import unittest
from typing import List
from unittest.mock import MagicMock, patch

import yaml
from click.testing import CliRunner

from data_pipelines_cli.cli import _cli
from data_pipelines_cli.cli_commands.compile import compile_project
from data_pipelines_cli.errors import DataPipelinesError, DockerNotInstalledError

goldens_dir_path = pathlib.Path(__file__).parent.parent.joinpath("goldens")


class CompileCommandTestCase(unittest.TestCase):
    @staticmethod
    def _execution_env_content(repository_url: str, tag: str):
        return {
            "image": {
                "repository": repository_url,
                "tag": tag,
            },
            "type": "k8s",
        }

    def setUp(self) -> None:
        self.all_subprocess_run_args = []

    def _mock_run(self, args: List[str]):
        self.all_subprocess_run_args += args

    @patch("pathlib.Path.cwd", lambda: goldens_dir_path)
    @patch("data_pipelines_cli.data_structures.git_revision_hash")
    def test_no_args(self, mock_git_revision_hash):
        commit_sha = "aaa9876aaa"
        mock_git_revision_hash.return_value = commit_sha

        runner = CliRunner()
        with tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_commands.compile.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch("data_pipelines_cli.config_generation.BUILD_DIR", pathlib.Path(tmp_dir)), patch(
            "data_pipelines_cli.cli_constants.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.dbt_utils.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.dbt_utils.subprocess_run", self._mock_run
        ):
            result = runner.invoke(_cli, ["compile"])
            self.assertEqual(0, result.exit_code, msg=result.exception)

            # 4 = 2 from goldens/dag, manifest.json, 'config' directory
            self.assertEqual(4, len(os.listdir(pathlib.Path(tmp_dir).joinpath("dag"))))

            args_str = " ".join(self.all_subprocess_run_args)
            self.assertIn("dbt deps", args_str)
            self.assertIn("dbt compile", args_str)
            self.assertIn("dbt docs generate", args_str)
            self.assertIn("dbt source freshness", args_str)

            tmp_dir_path = pathlib.Path(tmp_dir)
            with open(tmp_dir_path.joinpath("dag", "manifest.json"), "r") as tmp_manifest, open(
                goldens_dir_path.joinpath("target", "manifest.json"), "r"
            ) as golden_manifest:
                self.assertDictEqual(json.load(golden_manifest), json.load(tmp_manifest))
            with open(
                tmp_dir_path.joinpath("dag", "config", "base", "datahub.yml"), "r"
            ) as tmp_datahub, open(
                goldens_dir_path.joinpath("config", "base", "datahub.yml")
            ) as golden_datahub:
                self.assertDictEqual(yaml.safe_load(golden_datahub), yaml.safe_load(tmp_datahub))
            with open(
                tmp_dir_path.joinpath("dag", "config", "base", "execution_env.yml"), "r"
            ) as tmp_k8s:
                self.assertDictEqual(
                    self._execution_env_content("my_docker_repository_uri", "aaa9876aaa"),
                    yaml.safe_load(tmp_k8s),
                )

    @patch("pathlib.Path.cwd", lambda: goldens_dir_path)
    @patch("data_pipelines_cli.data_structures.git_revision_hash")
    def test_docker_not_installed(self, mock_git_revision_hash):
        commit_sha = "aaa9876aaa"
        mock_git_revision_hash.return_value = commit_sha

        runner = CliRunner()
        with patch.dict(
            "sys.modules", docker=None
        ), tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_commands.compile.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.config_generation.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.cli_constants.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.dbt_utils.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.dbt_utils.subprocess_run", self._mock_run
        ):
            result = runner.invoke(
                _cli,
                ["compile", "--docker-build"],
            )
        self.assertEqual(1, result.exit_code)
        self.assertIsInstance(result.exception, DockerNotInstalledError)

    @patch("pathlib.Path.cwd", lambda: goldens_dir_path)
    @patch("data_pipelines_cli.data_structures.git_revision_hash")
    def test_docker_uri_build(self, mock_git_revision_hash):
        commit_sha = "aaa9876aaa"
        mock_git_revision_hash.return_value = commit_sha

        docker_tag = None

        def _mock_docker(**kwargs):
            nonlocal docker_tag
            docker_tag = kwargs["tag"]
            return None, []

        docker_images_mock = MagicMock()
        docker_images_mock.configure_mock(**{"build": _mock_docker})
        docker_client_mock = MagicMock()
        docker_client_mock.configure_mock(**{"images": docker_images_mock})
        docker_mock = MagicMock()
        docker_mock.configure_mock(**{"from_env": lambda: docker_client_mock})

        runner = CliRunner()
        with patch.dict(
            "sys.modules", **{"docker": docker_mock, "docker.errors": MagicMock()}
        ), tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_commands.compile.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.cli_constants.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.config_generation.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.dbt_utils.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.dbt_utils.subprocess_run", self._mock_run
        ):
            result = runner.invoke(_cli, ["compile", "--docker-build"])
            self.assertEqual(0, result.exit_code, msg=result.exception)
            self.assertEqual("my_docker_repository_uri:aaa9876aaa", docker_tag)

    @patch("pathlib.Path.cwd", lambda: goldens_dir_path)
    @patch("data_pipelines_cli.data_structures.git_revision_hash")
    def test_docker_throw_build_error(self, mock_git_revision_hash):
        commit_sha = "aaa9876aaa"
        mock_git_revision_hash.return_value = commit_sha

        class MockException(Exception):
            msg = "some error message"
            build_log = ["more", "errors"]

        docker_errors_mock = MagicMock(BuildError=MockException)
        docker_images_mock = MagicMock(build=MagicMock(side_effect=MockException))
        docker_client_mock = MagicMock(images=docker_images_mock)
        docker_mock = MagicMock(from_env=lambda: docker_client_mock, errors=docker_errors_mock)

        with patch.dict(
            "sys.modules", **{"docker": docker_mock, "docker.errors": docker_errors_mock}
        ), tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_commands.compile.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.cli_constants.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.config_generation.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.dbt_utils.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.dbt_utils.subprocess_run", self._mock_run
        ):
            with self.assertRaises(DataPipelinesError):
                try:
                    compile_project("base", True)
                except MockException:
                    self.fail()

    @patch("pathlib.Path.cwd", lambda: goldens_dir_path)
    @patch("data_pipelines_cli.data_structures.git_revision_hash")
    def test_docker_throw_on_error(self, mock_git_revision_hash):
        commit_sha = "aaa9876aaa"
        mock_git_revision_hash.return_value = commit_sha

        def _mock_docker(**_kwargs):
            return None, [
                '{"status":"Build image"}',
                '{"errorDetail":{"message":"An image cannot be built."},'
                '"error":"An image cannot be built."}',
            ]

        class MockException(Exception):
            pass

        docker_errors_mock = MagicMock()
        docker_errors_mock.BuildError = MockException

        docker_images_mock = MagicMock()
        docker_images_mock.configure_mock(**{"build": _mock_docker})
        docker_client_mock = MagicMock()
        docker_client_mock.configure_mock(**{"images": docker_images_mock})
        docker_mock = MagicMock()
        docker_mock.configure_mock(
            **{"from_env": lambda: docker_client_mock, "errors": docker_errors_mock}
        )

        with patch.dict(
            "sys.modules", **{"docker": docker_mock, "docker.errors": docker_errors_mock}
        ), tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_commands.compile.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.cli_constants.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.config_generation.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.dbt_utils.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.dbt_utils.subprocess_run", self._mock_run
        ):
            with self.assertRaises(DataPipelinesError):
                compile_project("base", True)

    @patch("data_pipelines_cli.data_structures.git_revision_hash")
    def test_datahub_variables(self, mock_git_revision_hash):
        commit_sha = "aaa9876aaa"
        mock_git_revision_hash.return_value = commit_sha

        datahub_content = {
            "datahub_path": "{{ var('datahub_path') }}",
            "heathers": {"api-key": "{{ env_var('SECRET_KEY') }}"},
        }

        runner = CliRunner()
        with tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_commands.compile.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch("data_pipelines_cli.config_generation.BUILD_DIR", pathlib.Path(tmp_dir)), patch(
            "data_pipelines_cli.cli_constants.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.dbt_utils.BUILD_DIR", pathlib.Path(tmp_dir)
        ), patch(
            "data_pipelines_cli.dbt_utils.subprocess_run", self._mock_run
        ), patch.dict(
            "os.environ", SECRET_KEY="very_secret_keeeeeeeeeeeeeeeeeey.abcd"
        ), tempfile.TemporaryDirectory() as tmp_dir_2, patch(
            "pathlib.Path.cwd", lambda: pathlib.Path(tmp_dir_2)
        ):
            for dir in ["config", "dag", "target"]:
                shutil.copytree(
                    goldens_dir_path.joinpath(dir),
                    pathlib.Path(tmp_dir_2).joinpath(dir),
                )
            with open(
                pathlib.Path(tmp_dir_2).joinpath("config", "base", "datahub.yml"), "w"
            ) as datahub_file:
                yaml.dump(datahub_content, datahub_file)

            result = runner.invoke(_cli, ["compile", "--env", "datahub"])
            self.assertEqual(0, result.exit_code, msg=result.exception)

            tmp_dir_path = pathlib.Path(tmp_dir)

            expected_dict = {
                "datahub_path": "http://example.com/datahub/RaNdOmTe$T__PAHT",
                "heathers": {"api-key": "very_secret_keeeeeeeeeeeeeeeeeey.abcd"},
            }

            with open(
                tmp_dir_path.joinpath("dag", "config", "base", "datahub.yml"), "r"
            ) as tmp_datahub:
                self.assertDictEqual(expected_dict, yaml.safe_load(tmp_datahub))



from __future__ import annotations

import os
import pathlib
from typing import Dict, Set, Union

import fsspec
from fsspec import AbstractFileSystem

from .cli_utils import echo_subinfo
from .errors import DataPipelinesError


class LocalRemoteSync:
    """Synchronizes local directory with a cloud storage's one."""

    local_fs: AbstractFileSystem
    """FS representing local directory"""
    local_path_str: str
    """Path to local directory"""
    remote_path_str: str
    """Path/URI of the cloud storage directory"""
    _local_directory_suffixes: Set[str]

    def __init__(
        self,
        local_path: Union[str, os.PathLike[str]],
        remote_path: str,
        remote_kwargs: Dict[str, str],
    ) -> None:
        if not pathlib.Path(local_path).exists():
            raise DataPipelinesError(f"{local_path} does not exists. Run 'dp compile' before.")

        self.local_path_str = str(local_path).rstrip("/")
        self.local_fs = fsspec.filesystem("file")
        self.remote_fs, self.remote_path_str = fsspec.core.url_to_fs(
            remote_path.rstrip("/"), **remote_kwargs
        )
        self._local_directory_suffixes = set()

    def sync(self, delete: bool = True) -> None:
        """
        Send local files to the remote directory and (optionally) delete
        unnecessary ones.

        :param delete: Whether to delete remote files that are \
        no longer present in local directory
        :type delete: bool
        """
        self._push_sync()
        if delete:
            self._delete()

    def _push_sync(self) -> None:
        """Push every file to the remote."""

        # TODO: Is it "lazy" (checking what to update) or not?
        local_directory = self.local_fs.find(self.local_path_str)
        self._local_directory_suffixes = set()
        for local_file in local_directory:
            local_file_suffix = local_file[len(self.local_path_str) :]
            self._local_directory_suffixes.add(local_file_suffix)
            remote_path_with_suffix = self.remote_path_str + local_file_suffix
            echo_subinfo(f"- Pushing {str(local_file)} to {remote_path_with_suffix}")
            self.remote_fs.put_file(local_file, remote_path_with_suffix)

    def _delete(self) -> None:
        """Remove every file from remote that's not local."""
        remote_directory = self.remote_fs.find(self.remote_path_str)
        for remote_file in remote_directory:
            remote_file_suffix = remote_file[len(self.remote_path_str) :]
            if remote_file_suffix not in self._local_directory_suffixes:
                self.remote_fs.rm(remote_file)



import pathlib
import tempfile
import unittest
from typing import List
from unittest.mock import patch

from click.testing import CliRunner

from data_pipelines_cli.cli import _cli


class CleanCommandTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.subprocess_run_args = []

    def _mock_run(self, args: List[str]):
        self.subprocess_run_args = args

    def test_clean(self):
        with patch("data_pipelines_cli.cli_commands.clean.subprocess_run", self._mock_run):
            runner = CliRunner()
            result = runner.invoke(_cli, ["clean"])
            self.assertEqual(0, result.exit_code, msg=result.exception)
            self.assertListEqual(["dbt", "clean"], self.subprocess_run_args)

    def test_clean_remove_dir(self):
        with patch(
            "data_pipelines_cli.cli_commands.clean.subprocess_run", self._mock_run
        ), tempfile.TemporaryDirectory() as tmp_dir, patch(
            "data_pipelines_cli.cli_commands.clean.BUILD_DIR", pathlib.Path(tmp_dir)
        ):
            assert pathlib.Path(tmp_dir).exists()
            runner = CliRunner()
            result = runner.invoke(_cli, ["clean"])
            self.assertEqual(0, result.exit_code, msg=result.exception)
            self.assertListEqual(["dbt", "clean"], self.subprocess_run_args)
            assert not pathlib.Path(tmp_dir).exists()

            # Fix for Python 3.7.
            # Otherwise it throws, as there is no directory to remove.
            pathlib.Path(tmp_dir).mkdir(parents=True, exist_ok=True)



"""Utilities related to VCS."""

from copier.vcs import GIT_PREFIX


def add_suffix_to_git_template_path(template_path: str) -> str:
    """Add ``.git`` suffix to *template_path*, if necessary.

    Check if *template_path* starts with Git-specific prefix (e.g. `git://`),
    or `http://` or `https://` protocol. If so, then add ``.git`` suffix if
    not present. Does nothing otherwise (as *template_path* probably points to
    a local directory).

    :param template_path: Path or URI to Git-based repository
    :type template_path: str
    :return: *template_path* with ``.git`` as suffix, if necessary
    :rtype: str
    """
    if not template_path.startswith(GIT_PREFIX) and not template_path.startswith(
        ("http://", "https://")
    ):
        return template_path
    return template_path + ("" if template_path.endswith(".git") else ".git")



import os
from typing import Any, Dict

from jinja2 import BaseLoader, DictLoader
from jinja2.nativetypes import NativeEnvironment

from data_pipelines_cli.errors import JinjaVarKeyError


def _prepare_jinja_replace_environment(
    jinja_loader: BaseLoader, dbt_vars: Dict[str, Any]
) -> NativeEnvironment:
    def _jinja_vars(var_name: str) -> Any:
        return dbt_vars[var_name]

    def _jinja_env_vars(var_name: str) -> Any:
        return os.environ[var_name]

    jinja_env = NativeEnvironment(loader=jinja_loader)
    # Hacking Jinja to use our functions, following:
    # https://stackoverflow.com/a/6038550
    jinja_env.globals["var"] = _jinja_vars
    jinja_env.globals["env_var"] = _jinja_env_vars

    return jinja_env


def replace_vars_with_values(
    templated_dictionary: Dict[str, Any], dbt_vars: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Replace variables in given dictionary using Jinja template in its values.

    :param templated_dictionary: Dictionary with Jinja-templated values
    :type templated_dictionary: Dict[str, Any]
    :param dbt_vars: Variables to replace
    :type dbt_vars: Dict[str, Any]
    :return: Dictionary with replaced variables
    :rtype: Dict[str, Any]
    :raises JinjaVarKeyError: Variable referenced in Jinja template does not exist
    """
    jinja_loader = DictLoader(templated_dictionary)
    jinja_env = _prepare_jinja_replace_environment(jinja_loader, dbt_vars)

    rendered_settings = {}
    for setting_key, setting_old_value in templated_dictionary.items():
        if isinstance(setting_old_value, dict):
            rendered_settings[setting_key] = replace_vars_with_values(setting_old_value, dbt_vars)
        else:
            try:
                rendered_settings[setting_key] = jinja_env.get_template(setting_key).render()
            except TypeError:
                # Jinja is accepting only str or Template and fails on int, etc.
                rendered_settings[setting_key] = setting_old_value
            except KeyError as key_error:
                # Variable does not exist and _jinja_vars or _jinja_env_vars thrown
                raise JinjaVarKeyError(key_error.args[0])
    return rendered_settings



import pathlib

from data_pipelines_cli.data_structures import DataPipelinesConfig

#:
IMAGE_TAG_TO_REPLACE: str = "<IMAGE_TAG>"
#: Name of the environment and dbt target to use for a local machine
PROFILE_NAME_LOCAL_ENVIRONMENT = "local"
#: Name of the dbt target to use for a remote machine
PROFILE_NAME_ENV_EXECUTION = "env_execution"
AVAILABLE_ENVS = [PROFILE_NAME_LOCAL_ENVIRONMENT, PROFILE_NAME_ENV_EXECUTION]

#: Content of the config file created by `dp init` command if no template path
#: is provided
DEFAULT_GLOBAL_CONFIG: DataPipelinesConfig = {
    "templates": {},
    "vars": {},
}

ENV_CONFIGURATION_PATH: pathlib.Path = pathlib.Path.home().joinpath(".dp.yml")
BUILD_DIR: pathlib.Path = pathlib.Path.cwd().joinpath("build")


def get_dbt_profiles_env_name(env: str) -> str:
    """
    Given a name of the environment, returns one of target names expected by
    the `profiles.yml` file.

    :param env: Name of the environment
    :type env: str
    :return: Name of the `target` to be used in `profiles.yml`
    """
    return (
        PROFILE_NAME_LOCAL_ENVIRONMENT
        if env == PROFILE_NAME_LOCAL_ENVIRONMENT
        else PROFILE_NAME_ENV_EXECUTION
    )



import shutil

import click

from ..cli_constants import BUILD_DIR
from ..cli_utils import echo_info, echo_subinfo, subprocess_run


def _dbt_clean() -> None:
    echo_info("dbt clean")
    subprocess_run(["dbt", "clean"])


def _remove_build_dir() -> None:
    if BUILD_DIR.exists():
        echo_info(f"Removing {BUILD_DIR}")
        shutil.rmtree(BUILD_DIR)
        echo_subinfo(f"{BUILD_DIR} removed")


def clean() -> None:
    """Delete local working directories."""
    _dbt_clean()
    _remove_build_dir()


@click.command(name="clean", help="Delete local working directories")
def clean_command() -> None:
    clean()



import sys
from typing import Any, Dict, List

import yaml

from data_pipelines_cli.cli_utils import echo_warning
from data_pipelines_cli.errors import DataPipelinesError, NoConfigFileError
from data_pipelines_cli.io_utils import git_revision_hash

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict


class TemplateConfig(TypedDict):
    """
    POD representing value referenced in the `templates` section of
    the `.dp.yml` config file.
    """

    template_name: str
    """Name of the template"""
    template_path: str
    """Local path or Git URI to the template repository"""


class DataPipelinesConfig(TypedDict):
    """POD representing `.dp.yml` config file."""

    templates: Dict[str, TemplateConfig]
    """Dictionary of saved templates to use in `dp create` command"""
    vars: Dict[str, str]
    """Variables to be passed to dbt as `--vars` argument"""


def read_env_config() -> DataPipelinesConfig:
    """
    Parse `.dp.yml` config file, if it exists. Otherwise, raises
    :exc:`.NoConfigFileError`.

    :return: POD representing `.dp.yml` config file, if it exists
    :rtype: DataPipelinesConfig
    :raises NoConfigFileError: `.dp.yml` file not found
    """
    # Avoiding a dependency loop between `cli_constants` and `data_structures`
    from data_pipelines_cli.cli_constants import ENV_CONFIGURATION_PATH

    if not ENV_CONFIGURATION_PATH.is_file():
        echo_warning(
            "No configuration file found. Run 'dp init' to create it.",
        )
        raise NoConfigFileError()

    with open(ENV_CONFIGURATION_PATH, "r") as f:
        return yaml.safe_load(f)


class DockerArgs:
    """Arguments required by the Docker to make a push to the repository.

    :raises DataPipelinesError: *repository* variable not set or git hash not found
    """

    repository: str
    """URI of the Docker images repository"""
    commit_sha: str
    """Long hash of the current Git revision. Used as an image tag"""

    def __init__(self, env: str) -> None:
        self.repository = self._get_docker_repository_uri_from_k8s_config(env)
        commit_sha = git_revision_hash()
        if not commit_sha:
            raise DataPipelinesError("Could not get git revision hash.")
        self.commit_sha = commit_sha

    def docker_build_tag(self) -> str:
        """
        Prepare a tag for Docker Python API build command.

        :return: Tag for Docker Python API build command
        :rtype: str
        """
        return f"{self.repository}:{self.commit_sha}"

    @staticmethod
    def _get_docker_repository_uri_from_k8s_config(env: str) -> str:
        # Avoiding a dependency loop between `cli_constants` and `data_structures`
        from data_pipelines_cli.cli_constants import BUILD_DIR
        from data_pipelines_cli.config_generation import (
            read_dictionary_from_config_directory,
        )

        execution_env_config = read_dictionary_from_config_directory(
            BUILD_DIR.joinpath("dag"), env, "execution_env.yml"
        )
        try:
            return execution_env_config["image"]["repository"]
        except KeyError as key_error:
            raise DataPipelinesError(
                f"Could not find 'repository' variable in build/config/{env}/execution_env.yml."
            ) from key_error


class DbtTableColumn(TypedDict, total=False):
    """POD representing a single column from 'schema.yml' file."""

    name: str
    description: str
    meta: Dict[str, Any]
    quote: bool
    tests: List[str]
    tags: List[str]


class DbtModel(TypedDict, total=False):
    """POD representing a single model from 'schema.yml' file."""

    name: str
    description: str
    meta: Dict[str, Any]
    identifier: str
    tests: List[str]
    tags: List[str]
    columns: List[DbtTableColumn]


class DbtSource(TypedDict, total=False):
    """POD representing a single source from 'schema.yml' file."""

    name: str
    description: str
    database: str
    schema: str
    meta: Dict[str, Any]
    tags: List[str]
    tables: List[DbtModel]



"""data_pipelines_cli module."""

from setuptools import find_packages, setup

with open("README.md") as f:
    README = f.read()

INSTALL_REQUIREMENTS = [
    "MarkupSafe==2.0.1",  # https://markupsafe.palletsprojects.com/en/2.1.x/changes/#version-2-1-0
    "dbt-core==1.0.2",
    "click>=8.0.3,<9.0",
    "questionary==1.10.0",
    "pyyaml>=5.1, <6.0",
    "types-PyYAML>=6.0",
    "copier==5.1.0",
    "Jinja2>=2.11,<2.12",
    "fsspec",
    "packaging>=20.4,<21.0",
]

EXTRA_FILESYSTEMS_REQUIRE = {
    "gcs": ["gcsfs"],
    "s3": ["s3fs"],
}

EXTRA_REQUIRE = {
    "docker": ["docker>=5.0"],
    "datahub": ["acryl-datahub>=0.8.17, <0.8.18"],
    "git": ["GitPython==3.1.26"],
    "tests": [
        "pytest>=6.2.2, <7.0.0",
        "pytest-cov>=2.8.0, <3.0.0",
        "pre-commit==2.15.0",
        "tox==3.21.1",
        "moto[s3]==2.2.16",
        "gcp-storage-emulator==2021.12.2",
        "GitPython==3.1.26",
        *(
            [
                require
                for requires_list in EXTRA_FILESYSTEMS_REQUIRE.values()
                for require in requires_list
            ]
        ),
    ],
    "docs": [
        "sphinx==4.3.1",
        "sphinx-rtd-theme==1.0.0",
        "sphinx-click>=3.1,<3.2",
        "myst-parser>=0.16, <0.17",
        "GitPython==3.1.26",
    ],
    **EXTRA_FILESYSTEMS_REQUIRE,
}

setup(
    name="data_pipelines_cli",
    version="0.15.1",
    description="CLI for data platform",
    long_description=README,
    long_description_content_type="text/markdown",
    license="Apache Software License (Apache 2.0)",
    license_files=("LICENSE",),
    python_requires=">=3",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="dbt airflow cli",
    author=u"Andrzej Swatowski",
    author_email="andrzej.swatowski@getindata.com",
    url="https://github.com/getindata/data-pipelines-cli/",
    packages=find_packages(exclude=["docs", "tests"]),
    include_package_data=True,
    install_requires=INSTALL_REQUIREMENTS,
    extras_require=EXTRA_REQUIRE,
    entry_points={"console_scripts": ["dp=data_pipelines_cli.cli:cli"]},
)



import click

from ..config_generation import get_profiles_dir_build_path
from ..dbt_utils import run_dbt_command
from .compile import compile_project


def test(env: str) -> None:
    """
    Run tests of the project on the local machine.

    :param env: Name of the environment
    :type env: str
    """
    compile_project(env)
    profiles_path = get_profiles_dir_build_path(env)
    run_dbt_command(("test",), env, profiles_path)


@click.command(name="test", help="Run tests of the project on the local machine")
@click.option(
    "--env",
    default="local",
    type=str,
    show_default=True,
    help="Name of the environment",
)
def test_command(env: str) -> None:
    test(env)



import pathlib
import shutil

import click
import yaml

from ..cli_constants import BUILD_DIR, IMAGE_TAG_TO_REPLACE
from ..cli_utils import echo_info, echo_warning
from ..config_generation import (
    copy_config_dir_to_build_dir,
    copy_dag_dir_to_build_dir,
    generate_profiles_yml,
)
from ..data_structures import DockerArgs
from ..dbt_utils import read_dbt_vars_from_configs, run_dbt_command
from ..docker_response_reader import DockerResponseReader
from ..errors import DockerErrorResponseError, DockerNotInstalledError
from ..io_utils import replace
from ..jinja import replace_vars_with_values


def _docker_build(docker_args: DockerArgs) -> None:
    """
    :param docker_args: Arguments required by the Docker to make a push to \
        the repository
    :raises DataPipelinesError: Docker not installed
    """
    try:
        import docker
        import docker.errors
    except ModuleNotFoundError:
        raise DockerNotInstalledError()

    echo_info("Building Docker image")
    docker_client = docker.from_env()
    docker_tag = docker_args.docker_build_tag()
    try:
        _, logs_generator = docker_client.images.build(path=".", tag=docker_tag)
        DockerResponseReader(logs_generator).click_echo_ok_responses()
    except docker.errors.BuildError as err:
        build_log = "\n".join([str(log) for log in err.build_log])
        raise DockerErrorResponseError(f"{err.msg}\n{build_log}")


def _dbt_compile(env: str) -> None:
    profiles_path = generate_profiles_yml(env, False)
    echo_info("Running dbt commands:")
    run_dbt_command(("deps",), env, profiles_path)
    run_dbt_command(("compile",), env, profiles_path)
    run_dbt_command(("docs", "generate"), env, profiles_path)
    run_dbt_command(("source", "freshness"), env, profiles_path)


def _copy_dbt_manifest() -> None:
    echo_info("Copying DBT manifest")
    shutil.copyfile(
        pathlib.Path.cwd().joinpath("target", "manifest.json"),
        BUILD_DIR.joinpath("dag", "manifest.json"),
    )


def replace_image_settings(docker_args: DockerArgs) -> None:
    k8s_config = BUILD_DIR.joinpath("dag", "config", "base", "execution_env.yml")
    echo_info(f"Replacing {IMAGE_TAG_TO_REPLACE} with commit SHA = {docker_args.commit_sha}")
    replace(k8s_config, IMAGE_TAG_TO_REPLACE, docker_args.commit_sha)


def _replace_datahub_with_jinja_vars(env: str) -> None:
    datahub_config_path: pathlib.Path = BUILD_DIR.joinpath("dag", "config", "base", "datahub.yml")

    if not datahub_config_path.exists():
        echo_warning(
            f"File config/base/datahub.yml does not exist in {BUILD_DIR}. "
            "Content will not be replaced."
        )
        return

    echo_info(f"Replacing Jinja variables in {datahub_config_path}.")
    with open(datahub_config_path, "r") as datahub_config_file:
        updated_config = replace_vars_with_values(
            yaml.safe_load(datahub_config_file), read_dbt_vars_from_configs(env)
        )
    with open(datahub_config_path, "w") as datahub_config_file:
        yaml.dump(updated_config, datahub_config_file)


def compile_project(
    env: str,
    docker_build: bool = False,
) -> None:
    """
    Create local working directories and build artifacts.

    :param env: Name of the environment
    :type env: str
    :param docker_build: Whether to build a Docker image
    :type docker_build: bool
    :raises DataPipelinesError:
    """
    copy_dag_dir_to_build_dir()
    copy_config_dir_to_build_dir()

    docker_args = DockerArgs(env)
    replace_image_settings(docker_args)
    _replace_datahub_with_jinja_vars(env)

    _dbt_compile(env)
    _copy_dbt_manifest()

    if docker_build:
        _docker_build(docker_args)


@click.command(
    name="compile",
    help="Create local working directories and build artifacts",
)
@click.option(
    "--env",
    default="base",
    type=str,
    show_default=True,
    required=True,
    help="Name of the environment",
)
@click.option(
    "--docker-build",
    is_flag=True,
    default=False,
    help="Whether to build a Docker image",
)
def compile_project_command(
    env: str,
    docker_build: bool,
) -> None:
    compile_project(env, docker_build)



import io
import json
from typing import Any, Dict, Optional, cast

import click
import yaml

from ..cli_constants import BUILD_DIR
from ..cli_utils import echo_error, echo_info, subprocess_run
from ..config_generation import read_dictionary_from_config_directory
from ..data_structures import DockerArgs
from ..docker_response_reader import DockerResponseReader
from ..errors import (
    AirflowDagsPathKeyError,
    DataPipelinesError,
    DependencyNotInstalledError,
    DockerErrorResponseError,
    DockerNotInstalledError,
)
from ..filesystem_utils import LocalRemoteSync


class DeployCommand:
    """A class used to push and deploy the project to the remote machine."""

    docker_args: Optional[DockerArgs]
    """Arguments required by the Docker to make a push to the repository.
    If set to `None`, :meth:`deploy` will not make a push"""
    datahub_ingest: bool
    """Whether to ingest DataHub metadata"""
    blob_address_path: str
    """URI of the cloud storage to send build artifacts to"""
    provider_kwargs_dict: Dict[str, Any]
    """Dictionary of arguments required by a specific cloud storage provider,
    e.g. path to a token, username, password, etc."""

    def __init__(
        self,
        env: str,
        docker_push: bool,
        dags_path: Optional[str],
        provider_kwargs_dict: Optional[Dict[str, Any]],
        datahub_ingest: bool,
    ) -> None:
        self.docker_args = DockerArgs(env) if docker_push else None
        self.datahub_ingest = datahub_ingest
        self.provider_kwargs_dict = provider_kwargs_dict or {}

        try:
            self.blob_address_path = dags_path or read_dictionary_from_config_directory(
                BUILD_DIR.joinpath("dag"),
                env,
                "airflow.yml",
            )["dags_path"]
        except KeyError as key_error:
            raise AirflowDagsPathKeyError from key_error

    def deploy(self) -> None:
        """Push and deploy the project to the remote machine.

        :raises DependencyNotInstalledError: DataHub or Docker not installed
        :raises DataPipelinesError: Error while pushing Docker image
        """
        if self.docker_args:
            self._docker_push()

        if self.datahub_ingest:
            self._datahub_ingest()

        self._sync_bucket()

    def _docker_push(self) -> None:
        """
        :raises DockerNotInstalledError: Docker not installed
        :raises DataPipelinesError: Error while pushing Docker image
        """
        try:
            import docker
        except ModuleNotFoundError:
            raise DockerNotInstalledError()

        echo_info("Pushing Docker image")
        docker_client = docker.from_env()
        docker_args = cast(DockerArgs, self.docker_args)

        try:
            DockerResponseReader(
                docker_client.images.push(
                    repository=docker_args.repository,
                    tag=docker_args.commit_sha,
                    stream=True,
                    decode=True,
                )
            ).click_echo_ok_responses()
        except DockerErrorResponseError as err:
            echo_error(err.message)
            raise DataPipelinesError(
                "Error raised when pushing Docker image. Ensure that "
                "Docker image you try to push exists. Maybe try running "
                "'dp compile' first?"
            )

    @staticmethod
    def _datahub_ingest() -> None:
        """:raises DependencyNotInstalledError: DataHub not installed"""
        try:
            import datahub  # noqa: F401
        except ModuleNotFoundError:
            raise DependencyNotInstalledError("datahub")

        echo_info("Ingesting datahub metadata")
        subprocess_run(
            [
                "datahub",
                "ingest",
                "-c",
                str(BUILD_DIR.joinpath("dag", "config", "base", "datahub.yml")),
            ]
        )

    def _sync_bucket(self) -> None:
        echo_info("Syncing Bucket")
        LocalRemoteSync(
            BUILD_DIR.joinpath("dag"), self.blob_address_path, self.provider_kwargs_dict
        ).sync(delete=True)


@click.command(
    name="deploy",
    help="Push and deploy the project to the remote machine",
)
@click.option("--env", default="base", show_default=True, type=str, help="Name of the environment")
@click.option("--dags-path", required=False, help="Remote storage URI")
@click.option(
    "--blob-args",
    required=False,
    type=click.File("r"),
    help="Path to JSON or YAML file with arguments that should be passed to "
    "your Bucket/blob provider",
)
@click.option(
    "--docker-push",
    type=bool,
    is_flag=True,
    default=False,
    help="Whether to push image to the Docker repository",
)
@click.option(
    "--datahub-ingest",
    is_flag=True,
    default=False,
    help="Whether to ingest DataHub metadata",
)
def deploy_command(
    env: str,
    dags_path: Optional[str],
    blob_args: Optional[io.TextIOWrapper],
    docker_push: bool,
    datahub_ingest: bool,
) -> None:
    if blob_args:
        try:
            provider_kwargs_dict = json.load(blob_args)
        except json.JSONDecodeError:
            blob_args.seek(0)
            provider_kwargs_dict = yaml.safe_load(blob_args)
    else:
        provider_kwargs_dict = None

    DeployCommand(
        env,
        docker_push,
        dags_path,
        provider_kwargs_dict,
        datahub_ingest,
    ).deploy()
